package JavacrawlerGid;

import com.google.common.base.Predicate;
import org.openqa.selenium.*;
import org.openqa.selenium.chrome.ChromeDriver;
import org.openqa.selenium.phantomjs.PhantomJSDriver;
import org.openqa.selenium.phantomjs.PhantomJSDriverService;
import org.openqa.selenium.remote.DesiredCapabilities;
import org.openqa.selenium.support.ui.ExpectedCondition;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.concurrent.TimeUnit;

/**
 * Created by Kong_ on 07.03.2016.
 */
public class DriverManager {
    private WebDriver driver;
    protected static DesiredCapabilities sCaps;

    public DriverManager() {
        System.setProperty( "phantomjs.binary.path", "phantomjs-2.1.1-windows\\bin\\phantomjs.exe" );

        //System.setProperty("webdriver.chrome.driver", "F:\\Drive\\Skole\\Event Rec\\Scraper\\JavaScraper\\chromedriver.exe");

        // Prepare capabilities
        sCaps = new DesiredCapabilities();
        sCaps.setJavascriptEnabled(true);
        sCaps.setCapability("takesScreenshot", false);

        // Change "User-Agent" via page-object capabilities
        sCaps.setCapability(PhantomJSDriverService.PHANTOMJS_PAGE_SETTINGS_PREFIX + "userAgent", "My User Agent - Chrome");

        // Disable "web-security", enable all possible "ssl-protocols" and "ignore-ssl-errors" for PhantomJSDriver
        sCaps.setCapability(PhantomJSDriverService.PHANTOMJS_CLI_ARGS, new String[] {
                "--web-security=false",
                "--ssl-protocol=any",
                "--ignore-ssl-errors=true",
                "--webdriver-loglevel=NONE"
        });
        this.driver = new PhantomJSDriver(sCaps);
        this.driver.manage().timeouts().implicitlyWait(30, TimeUnit.SECONDS);
        //this.driver.manage().timeouts().setScriptTimeout(30, TimeUnit.SECONDS);
        this.driver.manage().timeouts().pageLoadTimeout(10, TimeUnit.SECONDS);

    }

    public void clickOnClassElement(String className) {
        WebElement element = this.driver.findElement(By.className(className));
        element.click();
    }

    public void waitForElement(String xpath) {
        /*this.driver.manage().timeouts().implicitlyWait(500, TimeUnit.MILLISECONDS);
        long end = System.currentTimeMillis() + 5000;
        while ( System.currentTimeMillis() < end ) {
            boolean elementIsPresent = this.driver.findElements(By.xpath(xpath)).size() != 0;
            if (elementIsPresent) break;
        }*/
        boolean loop = true;
        while(loop) {
            try {
                loop = !waitForJStoLoad();
            } catch (TimeoutException e) {
                System.out.println("Reloading");
                RenderURL(this.driver.getCurrentUrl());
            }
        }
        this.driver.manage().timeouts().implicitlyWait(30, TimeUnit.SECONDS);

    }

    public boolean waitForJStoLoad() throws TimeoutException {

        WebDriverWait wait = new WebDriverWait(driver, 30);

        // wait for jQuery to load
        ExpectedCondition<Boolean> jQueryLoad = new ExpectedCondition<Boolean>() {
            @Override
            public Boolean apply(WebDriver driver) {
                try {
                    return ((long)((JavascriptExecutor)driver).executeScript("return jQuery.active") == 0);
                }
                catch (Exception e) {
                    return true;
                }
            }
        };

        // wait for Javascript to load
        ExpectedCondition<Boolean> jsLoad = new ExpectedCondition<Boolean>() {
            @Override
            public Boolean apply(WebDriver driver) {
                return ((JavascriptExecutor)driver).executeScript("return document.readyState")
                        .toString().equals("complete");
            }
        };

        return wait.until(jQueryLoad) && wait.until(jsLoad);
    }

    public void Exit() {
        this.driver.quit();
    }

    public void LoadURL(String url) {
        this.driver.get(url);
    }

    public String getPageSource(){
        return this.driver.getPageSource();
    }

    public String RenderURLAndWaitForElement(String url, String xpath){
        RenderURL(url);
        waitForElement(xpath);
        return this.driver.getPageSource();
    }

    public String RenderURL(String url) {
        System.out.println("Render this: " + url);
        this.driver.get("http://example.com/");
        boolean loop = true;
        while(loop) {
            try {
                this.driver.get(url);
                loop = false;
            } catch (TimeoutException e) {
                System.out.println("Reloading");
            }
        }
        System.out.println("Renders: " + driver.getCurrentUrl());
        return this.driver.getPageSource();
    }

    public String getCurrentURL() {
        return this.driver.getCurrentUrl();
    }
}
