package JavacrawlerGid;

import org.apache.commons.io.FileUtils;
import org.htmlcleaner.CleanerProperties;
import org.htmlcleaner.HtmlCleaner;
import org.htmlcleaner.SimpleXmlSerializer;
import org.htmlcleaner.TagNode;
import org.json.JSONArray;
import org.json.JSONObject;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.*;
import java.io.*;
import java.net.URLEncoder;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.regex.Pattern;
import java.util.stream.Stream;

/**
 * Created by madslp on 08.03.2016.
 */
public class FilmwebScraper {

    private final Pattern pattern = Pattern.compile("^-?\\d+$");

    private DriverManager driverManager;
    private HtmlCleaner cleaner;
    private CleanerProperties props;
    private TagNode tagNode;
    private ByteArrayOutputStream out;
    private DocumentBuilderFactory factory;
    private DocumentBuilder builder;
    private Document doc;
    private XPathFactory xPathFactory;
    private XPath xpath;
    private XPathExpression xPathExpression;
    private String today;
    private Calendar calendar;

    private String cinemasFilename;
    private String moviesFilename;
    private ArrayList<String> screeningFilenames;
    private HashMap<String, JSONObject> locations;
    private HashMap<String, JSONObject> movies;
    private HashMap<String, JSONObject> screeningsMap;

    private ArrayList<JSONObject> needReload = new ArrayList<>();




    public FilmwebScraper() {

        eraseFile("NeedToReload.txt");

        driverManager = new DriverManager();

        Runtime.getRuntime().addShutdownHook(new Thread(() -> { driverManager.Exit(); }));

        cleaner = new HtmlCleaner();
        props = cleaner.getProperties();
        factory = DocumentBuilderFactory.newInstance();
        xPathFactory = XPathFactory.newInstance();

        calendar = new GregorianCalendar();
        today = LocalDateTime.now().format(DateTimeFormatter.ofPattern("dd.MM.yyyy"));

        cinemasFilename = "cinemasWithCoords.json";
        moviesFilename = "movies"+today+".json";
        screeningFilenames = new ArrayList<>();

        try {
            FileUtils.cleanDirectory(new File("screenings"));
        } catch (IOException e) {
            e.printStackTrace();
        }


        locations = readJSONObjectsFromFile(cinemasFilename);
        movies = readJSONObjectsFromFile(moviesFilename);


        for (Map.Entry<String, JSONObject> entry : locations.entrySet()) {
            JSONArray cinemas = entry.getValue().getJSONArray("containsPlace");
            String location = entry.getKey();
            String theater = "";
            for(int i = 0; i < cinemas.length(); i++){
                JSONObject cinema = cinemas.getJSONObject(i);
                if(cinemas.length() > 1){
                    theater = cinema.getString("@id");
                }
                ArrayList<String> availableDates = getDates(theater, location);
                for (String dateSearch : availableDates) {
                    String screenings = getScreenings(theater, location, dateSearch);
                    NodeList eventsByDay = getNodeList(screenings, "//div[@id='programContainer']/div/div");
                    scrapeEvents(eventsByDay, cinema, driverManager.getCurrentURL());
                }
            }
        }
        while(needReload.size() > 0) {
            JSONObject cinema = needReload.remove(needReload.size()-1);
            String location = cinema.getString("containedInPlace");
            String theater = cinema.getString("@id");
            ArrayList<String> availableDates = getDates(theater, location);
            for (String dateSearch : availableDates) {
                String screenings = getScreenings(theater, location, dateSearch);
                NodeList eventsByDay = getNodeList(screenings, "//div[@id='programContainer']/div/div");
                scrapeEvents(eventsByDay, cinema, driverManager.getCurrentURL());
            }
        }

        driverManager.Exit();
        System.out.println("CLOSING");
        System.exit(0);
    }

    public void scrapeEvents(NodeList eventsByDay, JSONObject cinema, String screeningURL) {
        boolean containerHasLoaded = false;
        System.out.println(driverManager.getCurrentURL());
        for(int r = 0; r < eventsByDay.getLength(); r++) {
            Node currentNode = eventsByDay.item(r);
            Node className = currentNode.getAttributes().getNamedItem("class");
            if(className != null && className.getTextContent().trim().equals("noShows row")) {
                System.out.println("No ScreeningEvents for this cinema: " + cinema.get("@id"));
                containerHasLoaded = true;
                continue;
            }

            screeningsMap = new HashMap();
            Node dateNode = currentNode.getFirstChild();
            currentNode.removeChild(dateNode);
            String date = getDateFromHeaderText(dateNode.getTextContent().trim());
            NodeList articles = currentNode.getChildNodes(); //Articles is all but first
            for(int a = 0; a < articles.getLength(); a++){
                Node textColumnNode = articles.item(a).getLastChild();
                Node textRowNode = textColumnNode.getFirstChild();
                Node movieInfoNode = textRowNode.getLastChild();
                Node movieInfoDataNode = movieInfoNode.getFirstChild();
                Node movieTitleNode = movieInfoDataNode.getFirstChild();
                if(!movieTitleNode.getAttributes().getNamedItem("class").getTextContent().trim().equals("movieTitle")){ //Need to remove a 50% off offer from the movie club that might occur
                    movieInfoDataNode.removeChild(movieTitleNode);
                    movieTitleNode = movieInfoDataNode.getFirstChild();
                }

                Node movieLinkNode = movieTitleNode.getFirstChild();

                Node descriptionNode = movieInfoDataNode.getChildNodes().item(2);
                Node descriptionTextNode = descriptionNode.getFirstChild();

                Node offerNode = textColumnNode.getLastChild();
                Node offerDataNode = offerNode.getFirstChild();

                JSONObject screening;

                String name = movieTitleNode.getTextContent().trim();
                String image = null;
                if (!articles.item(a).getFirstChild().getFirstChild().getAttributes().getNamedItem("class").getTextContent().equals("noPoster")) {
                    image = articles.item(a).getFirstChild().getFirstChild().getFirstChild().getAttributes().getNamedItem("src").getTextContent();
                }
                String workPresented = null;
                if(movieLinkNode != null && movieLinkNode.getAttributes() != null && movieLinkNode.getAttributes().getNamedItem("href") != null){
                    workPresented = movieLinkNode.getAttributes().getNamedItem("href").getTextContent().trim();
                    if (!movies.containsKey(workPresented)) getMetaDataFromMovie(workPresented);
                }

                String description = descriptionTextNode.getTextContent().trim();

                for (int j = 0; j < offerDataNode.getChildNodes().getLength(); j++) {
                    Node purchase = offerDataNode.getChildNodes().item(j);


                    String location = purchase.getChildNodes().item(0).getTextContent().trim();
                    Node showStart = purchase.getChildNodes().item(1);
                    String videoFormat = null;
                    String startDate;
                    if (showStart.getFirstChild().getAttributes().getNamedItem("class").getTextContent().equals("time")) {
                        startDate = showStart.getFirstChild().getTextContent().trim();
                    } else {
                        videoFormat = showStart.getFirstChild().getTextContent().trim();
                        startDate = showStart.getFirstChild().getNextSibling().getTextContent().trim();
                    }
                    String subtitleLanguage = purchase.getChildNodes().item(2).getTextContent().trim();

                    screening = new JSONObject();
                    screening.put("@context", "http://schema.org");
                    screening.put("@type", "ScreeningEvent");
                    screening.put("url", screeningURL);
                    screening.put("location", cinema);

                        /*
                        Orig. tale
                        Norsk tale
                        utekstet
                         */
                    if(workPresented != null) {

                        if (subtitleLanguage.contains("Norsk tale")) {
                            screening.put("inLanguage", "nb-NO");
                        } else { //if (subtitleLanguage.contains("Orig. tale")) {
                            if(movies.get(workPresented).has("inLanguage")) {
                                screening.put("inLanguage", movies.get(workPresented).get("inLanguage"));
                            } else if(movies.get(workPresented).has("countryOfOrigin")){
                                screening.put("inLanguage", movies.get(workPresented).getJSONObject("countryOfOrigin").get("name"));
                            } else {
                                System.out.println(movies.get(workPresented).get("name"));
                            }
                        }
                        if (!subtitleLanguage.contains("utekstet") || subtitleLanguage.contains("Norsk tale")) {
                            screening.put("subtitleLanguage", "nb-NO");
                        }
                        screening.put("workPresented", workPresented);
                        if(movies.get(workPresented).has("duration")){
                            screening.put("duration", movies.get(workPresented).get("duration"));
                        }
                        if(movies.get(workPresented).has("typicalAgeRange")) {
                            screening.put("typicalAgeRange", movies.get(workPresented).get("typicalAgeRange"));
                        }
                    }
                        /*
                        version v2d
                        version v3d
                         */
                    if (videoFormat != null) screening.put("videoFormat", videoFormat);
                    if(!location.equals(cinema.getString("@id"))){
                        JSONObject locationObject = new JSONObject();
                        locationObject.put("@type", "Place");
                        locationObject.put("name", location);
                        locationObject.put("containedInPlace", cinema);
                        screening.put("location", locationObject);

                    }
                    if (purchase.getAttributes().getNamedItem("href") != null) {
                        String offer = purchase.getAttributes().getNamedItem("href").getTextContent().trim();
                        screening.put("offer", offer);
                    }
                    screening.put("startDate", parseDateTimeToISO8601(date.split("\\."), startDate));
                    screening.put("description", description);
                    if (image != null) screening.put("image", image);
                    screening.put("name", name);

                    screeningsMap.put(UUID.randomUUID().toString(), screening);
                    containerHasLoaded = true;
                }
            }
            writeScreeningsToFile(date);
        }
        if(!containerHasLoaded){
            writeToFile("NeedToReload.txt", "Script not loaded for: " + cinema.getString("@id") + " : " + driverManager.getCurrentURL());
            if(!cinema.getString("@id").equals("Store Auditorium")){
                needReload.add(cinema);
            }
        }
    }

    public ArrayList<String> getDates(String theatre, String location) {
        ArrayList<String> dates = new ArrayList<>();
        String html = getScreenings(theatre, location, today);
        NodeList dateNodes = getNodeList(html, "//select[@class='dropdown dateOptions']/option/@value");
        if(dateNodes.getLength() > 0) {
            for(int i = 0; i < dateNodes.getLength(); i++){
                dates.add(dateNodes.item(i).getTextContent().trim());
            }
        }else {
            dates.add(today);
        }
        System.out.println(dates);
        return dates;
    }

    public String getDateFromHeaderText(String text) {
        String[] split = text.split(" ");
        String date = split.length == 1 ? split[0] : split[1];
        int day = Integer.parseInt(date.split("\\.")[0]);
        int month = Integer.parseInt(date.split("\\.")[1]);
        int year = calendar.get(Calendar.YEAR);
        if(month < calendar.get(Calendar.MONTH) ||
                (month == calendar.get(Calendar.MONTH) && day < calendar.get(Calendar.DAY_OF_MONTH)) )
        {
            year++;
        }
        return day + "." + month + "." + year;
    }

    private void getMetaDataFromMovie(String screeningUrl) {
        String html = driverManager.RenderURLAndWaitForElement(screeningUrl, "//div[@class='userAvg']");;
        System.out.println(screeningUrl);
        //System.out.println(html);
        JSONObject movie = new JSONObject();
        movies.put(screeningUrl, movie);
        movie.put("@context", "http://schema.org");
        movie.put("@type", "Movie");
        movie.put("@id", screeningUrl);
        movie.put("url", screeningUrl);

        ArrayList<String> descriptions = new ArrayList();


        ArrayList alternateNames = new ArrayList();
        Node headTitle = getNode(html, "//div[@class='filmomtale full']/h1");
        if(headTitle != null) alternateNames.add(headTitle.getTextContent().trim());
        Node largeTitle = getNode(html, "//div[@class='largeTitle']");
        if(headTitle != null)alternateNames.add(largeTitle.getTextContent().trim());

        Node titleNode = getNode(html, "//meta[@name='title']/@content"); //title
        String name = titleNode.getTextContent();
        movie.put("name", name);

        Node descriptionNode = getNode(html, "//meta[@name='description']/@content"); //description
        descriptions.add(descriptionNode.getTextContent());

        Node imageNode = getNode(html, "//meta[@property='og:image']/@content"); //description
        if(imageNode != null) {
            String image = imageNode.getTextContent();
            movie.put("image", image);
        }

        NodeList rating = getNodeList(html, "//div[@class='userAvg']/span"); //Rating
        String ratingValue = rating.item(0).getTextContent().trim();
        String bestRating = rating.item(1).getTextContent().replaceAll("\\D+","");
        NodeList ratingMeta = getNode(html, "//div[@class='userAvg']").getParentNode().getChildNodes(); //RatingCount
        String ratingCount = "0";
        for(int i = 0; i < ratingMeta.getLength(); i++) {
            String text = ratingMeta.item(i).getTextContent();
            if(!ratingMeta.item(i).getNodeName().equals("script") && ratingMeta.item(i).getTextContent().contains("stemmer")) {
                ratingCount = text.replaceAll("\\D+","");
                break;
            }
        }

        HashMap aggregateRating = new HashMap();
        aggregateRating.put("@type", "AggregateRating");
        aggregateRating.put("bestRating", bestRating);
        aggregateRating.put("ratingCount", ratingCount);
        aggregateRating.put("ratingValue", ratingValue);
        movie.put("aggregateRating", aggregateRating);

        Node ingressNode = getNode(html, "//div[@class='ingress']"); //Om filmen ingress
        if(ingressNode != null){
            String ingress = ingressNode.getTextContent().trim();
            descriptions.add(ingress);
        }

        Node bodytextNode = getNode(html, "//div[@class='bodytext']"); //Om filmen bodytext
        if(bodytextNode != null) {
            String bodyText = bodytextNode.getTextContent().trim();
            descriptions.add(bodyText);
        }


        NodeList completeFacts = getNodeList(html, "//ul[@class='facts completeFacts']/li"); //Fakta Om filmen

        for(int k = 0; k < completeFacts.getLength(); k++) {
            Node firstChild = completeFacts.item(k).getFirstChild();
            String fact = firstChild.getTextContent().trim();
            completeFacts.item(k).removeChild(firstChild);
            NodeList textNodes = completeFacts.item(k).getChildNodes();
            switch(fact){
                case "Norgespremiere:":
                    String datePublished = parseDateToISO8601(textNodes.item(0).getTextContent().trim().split("\\s|\\."));
                    movie.put("datePublished", datePublished);
                    break;
                case "Originaltittel:":
                    alternateNames.add(textNodes.item(0).getTextContent().trim());
                    break;
                case "Genre:":
                    String[] genres = textNodes.item(0).getTextContent().trim().split(" / |, ");
                    movie.put("genre", genres);
                    break;
                case "Skuespillere:":
                    ArrayList actors = getPersonListWithUrl(textNodes);
                    movie.put("actor", actors);
                    break;
                case "Regi:":
                    ArrayList directors = getPersonListWithUrl(textNodes);
                    movie.put("director", directors);
                    break;
                case "Manus:":
                    ArrayList authors = getPersonList(textNodes.item(0).getTextContent().trim().split(", "));
                    movie.put("author", authors);
                    break;
                case "Nasjonalitet:":
                    HashMap countryOfOrigin = new HashMap();
                    countryOfOrigin.put("@type", "Country");
                    countryOfOrigin.put("name", textNodes.item(0).getTextContent().trim());
                    movie.put("countryOfOrigin", countryOfOrigin);
                    break;
                case "Språk:": //Should use IETF BCP 47 standard
                    movie.put("inLanguage", textNodes.item(0).getTextContent().trim());
                    break;
                case "Musikk:":
                    ArrayList composer = getPersonList(textNodes.item(0).getTextContent().trim().split(", "));
                    movie.put("musicBy", composer);
                    break;
                case "Produksjonsselskap:":
                    ArrayList productionCompanies = getOrganizationList(textNodes.item(0).getTextContent().trim().split(", "));
                    movie.put("productionCompany", productionCompanies);
                    break;
                case "Distribusjon:": //Maybe provider or publisher?
                    break;
                case "Produksjonsår:": //only year
                    break;
                case "Lengde:": //Duration, might be only hours or minutes. Usually both e.g. "2 t. 26 min."
                    String[] duration = textNodes.item(0).getTextContent().replaceAll("\\D+"," ").trim().split(" ");
                    if(duration.length == 1){
                        if(textNodes.item(0).getTextContent().contains("t")) {
                            movie.put("duration", "PT" + duration[0] + "H");
                        } else if(textNodes.item(0).getTextContent().contains("min")) {
                            movie.put("duration", "PT" + duration[0] + "M");
                        }
                    } else {
                        movie.put("duration", "PT" + duration[0] + "H" + duration[1] + "M");
                    }
                    break;
                case "Aldersgrense:":
                    String typicalAgeRange = textNodes.item(0).getTextContent().trim();
                    if(typicalAgeRange.equals("Tillatt for alle")){
                        movie.put("typicalAgeRange", "0-");
                    } else {
                        movie.put("typicalAgeRange", typicalAgeRange.replaceAll("\\D+","")+"-");
                    }
                    break;
                case "Begrunnelse:": //typicalAgeRange described
                    break;
                case "Video distribusjon:": //Distributors of the video
                    break;
                case "Egnethet:": //Proposed audiences
                    break;
                case "Medvirkende:": //"Actors" in a documentary?
                    break;
                default:
                    System.out.println("New fact: " + fact);
            }
        }

        NodeList trailerNodes = getNodeList(html, "//div[@id='trailere']/div/a"); //Trailers
        ArrayList trailers = new ArrayList();
        for (int i = 0; i < trailerNodes.getLength();i++){
            String trailerUrl = trailerNodes.item(i).getAttributes().getNamedItem("href").getTextContent().trim();
            String trailerName = trailerNodes.item(i).getTextContent().trim();
            HashMap videoObject = new HashMap();
            videoObject.put("@type", "VideoObject");
            videoObject.put("url", "http:" + trailerUrl);
            videoObject.put("name", trailerName);
            trailers.add(videoObject);
        }

        NodeList reviewNodes = getNodeList(html, "//div[@id='anmeldelser']/a"); //Reviews
        ArrayList reviews = new ArrayList();
        for (int i = 0; i < reviewNodes.getLength();i++){
            String reviewUrl = reviewNodes.item(i).getAttributes().getNamedItem("href").getTextContent().trim();
            String reviewPublisherName = reviewNodes.item(i).getTextContent().trim();
            String reviewRatingValue = reviewNodes.item(i).getAttributes().getNamedItem("class").getTextContent().replaceAll("\\D+"," ").trim();
            HashMap reviewPublisher = new HashMap();
            reviewPublisher.put("@type", "Organization");
            reviewPublisher.put("name", reviewPublisherName);
            HashMap reviewRating = new HashMap();
            reviewRating.put("@type", "Rating");
            if(reviewRatingValue.split(" ").length == 2) {
                reviewRating.put("bestRating", reviewRatingValue.split(" ")[1]);
                reviewRating.put("ratingValue", reviewRatingValue.split(" ")[0]);
            } else {
                reviewRating.put("bestRating", "6");
                reviewRating.put("ratingValue", reviewRatingValue);
            }
            HashMap reviewObject = new HashMap();
            reviewObject.put("@type", "Review");
            reviewObject.put("url", reviewUrl);
            reviewObject.put("publisher", reviewPublisher);
            reviewObject.put("reviewRating", reviewRating);
            reviews.add(reviewObject);
        }

        movie.put("review", reviews.toArray());
        movie.put("trailer", trailers.toArray());
        movie.put("description", descriptions.toArray());
        movie.put("alternateName", alternateNames.toArray());
        writeToFile(moviesFilename, movie.toString());
    }

    public String parseDateToISO8601(String[] unParsedDates){
        ArrayList<String> months = new ArrayList<>(Arrays.asList("JANUAR", "FEBRUAR", "MARS", "APRIL", "MAI", "JUNI", "JULI", "AUGUST", "SEPTEMBER", "OKTOBER", "NOVEMBER", "DESEMBER"));
        int pos = 0;
        int day = -1;
        if(unParsedDates.length == 3) {
            day = Integer.parseInt(unParsedDates[pos++]);
        }
        int month;
        String monthString = unParsedDates[pos++];
        if(pattern.matcher(monthString).find()) { //if the month is a number
            month = Integer.parseInt(monthString);
        } else {
            month = months.indexOf(monthString.toUpperCase())+1;
        }
        int year = Integer.parseInt(unParsedDates[pos++]);
        return String.format("%04d", year)+"-"+String.format("%02d", month)+ (day != -1 ? "-"+String.format("%02d", day) : "");
    }

    public String parseDateTimeToISO8601(String[] unParsedDates, String time){
        return parseDateToISO8601(unParsedDates)+"T"+time;
    }

    public ArrayList getPersonListWithUrl(NodeList personNodes){
        ArrayList personList = new ArrayList();
        for (int i = 0; i < personNodes.getLength();i++){
            if(!personNodes.item(i).getNodeName().equals("a")) continue;
            String actorName = personNodes.item(i).getTextContent().trim();
            HashMap person = new HashMap();
            person.put("@type", "Person");
            String personUrl = personNodes.item(i).getAttributes().getNamedItem("href").getTextContent().trim();
            person.put("url", "http:" + personUrl);
            person.put("name", actorName);
            personList.add(person);
        }
        return personList;
    }

    public ArrayList getPersonList(String[] names){
        ArrayList personList = new ArrayList();
        for (int i = 0; i < names.length; i++){
            String authorName = names[i];
            HashMap person = new HashMap();
            person.put("@type", "Person");
            person.put("name", authorName);
            personList.add(person);
        }
        return personList;
    }

    public ArrayList getOrganizationList(String[] names) {
        ArrayList organizationList = new ArrayList();
        for (int i = 0; i < names.length; i++){
            String organizationName = names[i];
            HashMap organization = new HashMap();
            organization.put("@type", "Organization");
            organization.put("name", organizationName);
            organizationList.add(organization);
        }
        return organizationList;
    }

    public void writeScreeningsToFile(String date){
        String filename = "screenings/screenings-" + date + ".json";
        if(!screeningFilenames.contains(filename)){
            //eraseFile(filename);
            screeningFilenames.add(filename);
        }
        writeMapToFile(screeningsMap, filename);
    }

    public HashMap<String, JSONObject> getScreeningsFromFile(String date){
        String filename = "screenings-" + date + ".json";
        return readJSONObjectsFromFile(filename);
    }

    public void eraseAndWriteToFile(HashMap<String, JSONObject> map, String filename){
        eraseFile(filename);
        writeMapToFile(map, filename);
    }

    public void writeMapToFile(HashMap<String, JSONObject> map, String filename){
        for (Map.Entry<String, JSONObject> entry : map.entrySet()) {
            writeToFile(filename, entry.getValue().toString());
        }
    }

    public void eraseFile(String filePath){
        try {
            Files.deleteIfExists(Paths.get(filePath));
            Files.createFile(Paths.get(filePath));
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void writeToFile(String filePath, String text) {
        try(PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(filePath, true)))) {
            System.out.println(text);
            out.println(text);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private HashMap<String, JSONObject> readJSONObjectsFromFile(String filename) {
        HashMap map = new HashMap();
        try (Stream<String> stream = Files.lines(Paths.get(filename))) {
            stream.forEach(line ->map.put((new JSONObject(line)).getString("@id"), new JSONObject(line)));
        } catch (IOException e) {
            e.printStackTrace();
        }
        return map;
    }

    public String getScreenings(String theater, String location, String date) {
        System.out.println("---------------------------------------------------------------------");
        System.out.println("Theater: " + theater + " - Location: " + location + " - Date: " + date);
        try {
            theater = URLEncoder.encode(theater, "UTF-8").replace("+", "%20");
            location = URLEncoder.encode(location, "UTF-8").replace("+", "%20");
        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        }
        String url = "http://www.filmweb.no/program/#location=" + location + "&date=" + date + "&page=pageShowsForDay&theater=" + theater ;

        String html = driverManager.RenderURLAndWaitForElement(url, "//div[@id='programContainer']");
        return html;
    }

    public NodeList getNodeList(String html, String xPathExpressionString) {
        tagNode = cleaner.clean(html);
        out = new ByteArrayOutputStream();
        NodeList list = null;
        try {
            new SimpleXmlSerializer(props).writeToStream(tagNode, out);

            builder = factory.newDocumentBuilder();
            doc = builder.parse(new ByteArrayInputStream(out.toByteArray()));
            xpath = xPathFactory.newXPath();
            xPathExpression = xpath.compile(xPathExpressionString);

            list = (NodeList) xPathExpression.evaluate(doc, XPathConstants.NODESET);
        } catch (IOException | SAXException | ParserConfigurationException | XPathExpressionException e) {
            e.printStackTrace();
        }
        return list;
    }

    public Node getNode(String html, String xPathExpressionString) {
        tagNode = cleaner.clean(html);
        out = new ByteArrayOutputStream();
        Node node = null;
        try {
            new SimpleXmlSerializer(props).writeToStream(tagNode, out);

            builder = factory.newDocumentBuilder();
            doc = builder.parse(new ByteArrayInputStream(out.toByteArray()));
            xpath = xPathFactory.newXPath();
            xPathExpression = xpath.compile(xPathExpressionString);

            node = (Node) xPathExpression.evaluate(doc, XPathConstants.NODE);
        } catch (IOException | SAXException | XPathExpressionException | ParserConfigurationException e) {
            e.printStackTrace();
        }
        return node;
    }

}
